#![allow(dead_code)]

use std::{
    collections::{BTreeMap, BTreeSet},
    fmt::{Debug, Display},
    sync::Arc,
};

use crate::{
    buffers::{Buffer, BufferDesc},
    commands, def_monotonic_id,
    device::{self, DeviceOwned},
    images::{self, Image, ImageDesc},
    util::{self, Rgba, WithLifetime},
    SwapchainFrame,
};
use ash::vk;
use itertools::Itertools;
use petgraph::{
    graph::NodeIndex,
    visit::{EdgeRef, IntoNodeReferences, NodeRef},
};

def_monotonic_id!(pub GraphResourceId);

#[derive(Debug, Clone)]
pub enum GraphResourceDesc {
    Image(ImageDesc),
    Buffer(BufferDesc),
}

#[derive(Debug, PartialEq, Eq)]
pub enum GraphResource {
    Framebuffer(Arc<SwapchainFrame>),
    ImportedImage(Arc<Image>),
    ImportedBuffer(Arc<Buffer>),
    Image(Arc<Image>),
    Buffer(Buffer),
}

#[derive(Debug, Clone, Copy)]
pub enum LoadOp {
    Clear(Rgba),
    Load,
    DontCare,
}

#[derive(Debug, Clone, Copy)]
pub enum StoreOp {
    DontCare,
    Store,
}

pub struct RenderContext<'a> {
    pub device: device::Device,
    pub cmd: commands::SingleUseCommand,
    pub resources: &'a BTreeMap<GraphResourceId, GraphResource>,
}

impl RenderContext<'_> {
    pub fn get_image(&self, id: GraphResourceId) -> Option<&Arc<Image>> {
        self.resources.get(&id).and_then(|res| match res {
            GraphResource::ImportedImage(arc) => Some(arc),
            GraphResource::Image(image) => Some(image),
            GraphResource::Framebuffer(fb) => Some(&fb.image),
            _ => None,
        })
    }
    pub fn get_buffer(&self, id: GraphResourceId) -> Option<&Buffer> {
        self.resources.get(&id).and_then(|res| match res {
            GraphResource::ImportedBuffer(arc) => Some(arc.as_ref()),
            GraphResource::Buffer(buffer) => Some(buffer),
            _ => None,
        })
    }
}

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord)]
pub struct Access {
    pub stage: vk::PipelineStageFlags2,
    pub mask: vk::AccessFlags2,
    pub layout: Option<vk::ImageLayout>,
}

impl core::ops::BitOr for Access {
    type Output = Self;

    fn bitor(self, rhs: Self) -> Self::Output {
        //assert_eq!(self.layout, rhs.layout);
        Self {
            stage: self.stage | rhs.stage,
            mask: self.mask | rhs.mask,
            layout: self.layout.max(rhs.layout),
        }
    }
}

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord)]
pub struct AccessMask {
    pub stage: vk::PipelineStageFlags2,
    pub mask: vk::AccessFlags2,
}
impl AccessMask {
    fn undefined() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
        }
    }
    fn is_empty(&self) -> bool {
        self.stage.is_empty() && self.mask.is_empty()
    }
}
impl core::ops::BitOr for AccessMask {
    type Output = Self;

    fn bitor(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage | rhs.stage,
            mask: self.mask | rhs.mask,
        }
    }
}
impl core::ops::Not for AccessMask {
    type Output = Self;

    fn not(self) -> Self::Output {
        Self {
            stage: !self.stage,
            mask: !self.mask,
        }
    }
}

impl core::ops::BitAnd for AccessMask {
    type Output = Self;

    fn bitand(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage & rhs.stage,
            mask: self.mask & rhs.mask,
        }
    }
}

impl core::ops::BitXor for AccessMask {
    type Output = Self;

    fn bitxor(self, rhs: Self) -> Self::Output {
        Self {
            stage: self.stage ^ rhs.stage,
            mask: self.mask ^ rhs.mask,
        }
    }
}

impl Access {
    pub fn into_access_mask(&self) -> AccessMask {
        AccessMask {
            stage: self.stage,
            mask: self.mask,
        }
    }
    pub fn empty() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: None,
        }
    }
    pub fn undefined() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::UNDEFINED),
        }
    }
    /// Only use this for `Ord`!
    pub fn min() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::from_raw(u64::MAX),
            mask: vk::AccessFlags2::from_raw(u64::MAX),
            layout: None,
        }
    }
    /// Only use this for `Ord`!
    pub fn max() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::from_raw(u64::MAX),
            mask: vk::AccessFlags2::from_raw(u64::MAX),
            layout: Some(vk::ImageLayout::from_raw(i32::MAX)),
        }
    }
    pub fn general() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::GENERAL),
        }
    }
    pub fn transfer_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::TRANSFER,
            mask: vk::AccessFlags2::TRANSFER_READ,
            layout: Some(vk::ImageLayout::TRANSFER_SRC_OPTIMAL),
        }
    }
    pub fn transfer_write() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::TRANSFER,
            mask: vk::AccessFlags2::TRANSFER_WRITE,
            layout: Some(vk::ImageLayout::TRANSFER_DST_OPTIMAL),
        }
    }
    pub fn vertex_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::VERTEX_ATTRIBUTE_INPUT,
            mask: vk::AccessFlags2::VERTEX_ATTRIBUTE_READ,
            layout: None,
        }
    }
    pub fn index_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::INDEX_INPUT,
            mask: vk::AccessFlags2::INDEX_READ,
            layout: None,
        }
    }
    pub fn indirect_read() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::DRAW_INDIRECT,
            mask: vk::AccessFlags2::INDIRECT_COMMAND_READ,
            layout: None,
        }
    }
    pub fn color_attachment_read_only() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_READ,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn color_attachment_write_only() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_WRITE,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn color_attachment_read_write() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::COLOR_ATTACHMENT_OUTPUT,
            mask: vk::AccessFlags2::COLOR_ATTACHMENT_WRITE
                | vk::AccessFlags2::COLOR_ATTACHMENT_READ,
            layout: Some(vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL),
        }
    }
    pub fn present() -> Self {
        Self {
            stage: vk::PipelineStageFlags2::NONE,
            mask: vk::AccessFlags2::empty(),
            layout: Some(vk::ImageLayout::PRESENT_SRC_KHR),
        }
    }
}

pub type RecordFn = dyn FnOnce(&RenderContext) -> crate::Result<()> + Send;

pub struct PassDesc {
    // this pass performs `Access` read on `GraphResourceId`.
    // some `GraphResourceId` may occur multiple times.
    pub reads: Vec<(GraphResourceId, Access)>,
    // this pass performs `Access` write on `GraphResourceId`.
    // some `GraphResourceId` may occur multiple times.
    pub writes: Vec<(GraphResourceId, Access)>,
    pub record: Box<RecordFn>,
}

impl Default for PassDesc {
    fn default() -> Self {
        Self {
            reads: Default::default(),
            writes: Default::default(),
            record: Box::new(|_| Ok(())),
        }
    }
}

impl Debug for PassDesc {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("PassDesc")
            .field("reads", &self.reads)
            .field("write", &self.writes)
            .finish_non_exhaustive()
    }
}

def_monotonic_id!(pub RenderGraphPassId);

// Non-imported resources remain `RenderGraphResourceDesc`s because they may be
// able to be aliased.
// This should be dual to liveness/register allocation in a compiler.
// Dummy-impl is just allocating every resource_desc itself. 5head-impl is trying
// to find resource_descs which are eq, but whose liveness doesn't overlap.
#[derive(Debug)]
pub struct RenderGraph {
    resource_descs: BTreeMap<GraphResourceId, GraphResourceDesc>,
    resources: BTreeMap<GraphResourceId, GraphResource>,
    accesses: BTreeMap<GraphResourceId, Access>,
    pass_descs: Vec<PassDesc>,
    /// the rendergraph produces these resources. Any passes on which these
    /// outputs do not depend are pruned.
    outputs: Vec<GraphResourceId>,
}

impl RenderGraph {
    pub fn new() -> Self {
        Self {
            resource_descs: BTreeMap::new(),
            resources: BTreeMap::new(),
            pass_descs: Vec::new(),
            accesses: BTreeMap::new(),
            outputs: Vec::new(),
        }
    }

    pub fn add_resource(&mut self, desc: GraphResourceDesc) -> GraphResourceId {
        let id = GraphResourceId::new();
        self.resource_descs.insert(id, desc);
        self.accesses.insert(id, Access::undefined());
        id
    }
    pub fn mark_as_output(&mut self, id: GraphResourceId) {
        // TODO: dedup
        self.outputs.push(id);
    }
    pub fn import_resource(&mut self, res: GraphResource, access: Access) -> GraphResourceId {
        if let Some((&id, _)) = self
            .resources
            .iter()
            .find(|(_, resident)| &&res == resident)
        {
            id
        } else {
            let id = GraphResourceId::new();
            self.resources.insert(id, res);
            self.accesses.insert(id, access);
            id
        }
    }
    pub fn import_image(&mut self, image: Arc<Image>, access: Access) -> GraphResourceId {
        let res = GraphResource::ImportedImage(image);
        self.import_resource(res, access)
    }
    pub fn import_buffer(&mut self, buffer: Arc<Buffer>, access: Access) -> GraphResourceId {
        let res = GraphResource::ImportedBuffer(buffer);
        self.import_resource(res, access)
    }
    pub fn import_framebuffer(&mut self, frame: Arc<SwapchainFrame>) -> GraphResourceId {
        let id = GraphResourceId::new();
        self.resources.insert(id, GraphResource::Framebuffer(frame));
        self.mark_as_output(id);
        id
    }
    pub fn add_pass(&mut self, pass: PassDesc) {
        self.pass_descs.push(pass);
    }

    // https://blog.traverseresearch.nl/render-graph-101-f42646255636
    // https://github.com/EmbarkStudios/kajiya/blob/main/crates/lib/kajiya-rg/src/graph.rs
    // https://themaister.net/blog/2017/08/15/render-graphs-and-vulkan-a-deep-dive/
    pub fn resolve(
        &mut self,
        device: device::Device,
    ) -> crate::Result<WithLifetime<'_, commands::CommandList<commands::SingleUseCommand>>> {
        // create internal resources:
        for (&id, desc) in self.resource_descs.iter() {
            tracing::trace!("creating resource {id:?} with {desc:?}");
            match desc.clone() {
                GraphResourceDesc::Image(image_desc) => {
                    self.resources.insert(
                        id,
                        GraphResource::Image(Arc::new(Image::new(device.clone(), image_desc)?)),
                    );
                }
                GraphResourceDesc::Buffer(buffer_desc) => {
                    self.resources.insert(
                        id,
                        GraphResource::Buffer(Buffer::new(device.clone(), buffer_desc)?),
                    );
                }
            }
        }

        #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
        enum PassNode {
            First,
            Pass(usize),
            Last,
        }

        impl PassNode {
            fn into_node_idx(&self) -> u32 {
                match self {
                    PassNode::First => 0,
                    PassNode::Last => 1,
                    PassNode::Pass(i) => 2 + *i as u32,
                }
            }
            fn into_u32(&self, max_i: u32) -> u32 {
                match self {
                    PassNode::First => 0,
                    PassNode::Last => max_i + 1,
                    PassNode::Pass(i) => 1 + *i as u32,
                }
            }
            fn range_full(from: Self, to: Self, max_i: u32) -> std::ops::RangeInclusive<u32> {
                from.into_u32(max_i)..=to.into_u32(max_i)
            }
            fn from_u32(v: u32, max_i: u32) -> Self {
                match v {
                    0 => Self::First,
                    n if n == 1 + max_i => Self::Last,
                    n => Self::Pass(n as usize - 1),
                }
            }
        }

        #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
        enum RefAccess {
            __Min,
            None,
            Read(Access),
            Write(Access),
            __Max,
        }

        #[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
        struct GraphRef {
            pass: PassNode,
            resource: GraphResourceId,
            access: RefAccess,
        }

        // gather references to resources.
        let (references, intervals) = {
            let mut references = BTreeSet::<GraphRef>::new();

            // interval for each resource from (first pass referencing, last pass referencing)
            let mut intervals = BTreeMap::<GraphResourceId, (PassNode, PassNode)>::new();

            // the root node creates and transitions resources added to the
            // graph. don't want to create any resources which are never used
            // after `First`. a newly created resource has no layout and no
            // writes to make-available.
            for rid in self.resource_descs.keys() {
                references.insert(GraphRef {
                    pass: PassNode::First,
                    resource: *rid,
                    access: RefAccess::Write(Access::undefined()),
                });
            }
            for rid in self.resources.keys() {
                references.insert(GraphRef {
                    pass: PassNode::First,
                    resource: *rid,
                    access: RefAccess::Write(
                        self.accesses
                            .get(rid)
                            .cloned()
                            .unwrap_or(Access::undefined()),
                    ),
                });

                intervals.insert(*rid, (PassNode::First, PassNode::First));
            }

            for (i, pass) in self.pass_descs.iter().enumerate() {
                let mut reads = BTreeMap::new();
                for (rid, access) in &pass.reads {
                    reads
                        .entry(*rid)
                        .and_modify(|entry| {
                            *entry = *entry | *access;
                        })
                        .or_insert(*access);

                    intervals
                        .entry(*rid)
                        .and_modify(|entry| {
                            entry.1 = PassNode::Pass(i);
                        })
                        .or_insert((PassNode::Pass(i), PassNode::Pass(i)));
                }
                references.extend(reads.into_iter().map(|(resource, access)| GraphRef {
                    pass: PassNode::Pass(i),
                    resource,
                    access: RefAccess::Read(access),
                }));

                let mut writes = BTreeMap::new();
                for (rid, access) in &pass.writes {
                    writes
                        .entry(*rid)
                        .and_modify(|entry| {
                            *entry = *entry | *access;
                        })
                        .or_insert(*access);

                    intervals
                        .entry(*rid)
                        .and_modify(|entry| {
                            entry.1 = PassNode::Pass(i);
                        })
                        .or_insert((PassNode::Pass(i), PassNode::Pass(i)));
                }

                references.extend(writes.into_iter().map(|(resource, access)| GraphRef {
                    pass: PassNode::Pass(i),
                    resource,
                    access: RefAccess::Write(access),
                }));
            }

            // any resource marked as output should be created and returned even
            // if it isn't referenced by any pass.
            for rid in &self.outputs {
                references.insert(GraphRef {
                    pass: PassNode::Last,
                    resource: *rid,
                    access: RefAccess::None,
                });

                intervals
                    .entry(*rid)
                    .and_modify(|entry| {
                        entry.1 = PassNode::Last;
                    })
                    .or_insert((PassNode::Last, PassNode::Last));
            }

            eprintln!("references: {references:#?}");
            eprintln!("intervals: {intervals:#?}");

            (references, intervals)
        };

        #[derive(Debug)]
        enum Barrier {
            Logical,
            Execution {
                src: vk::PipelineStageFlags2,
                dst: vk::PipelineStageFlags2,
            },
            LayoutTransition {
                src: (vk::PipelineStageFlags2, vk::ImageLayout),
                dst: (vk::PipelineStageFlags2, vk::ImageLayout),
            },

            MakeAvailable {
                src: (vk::PipelineStageFlags2, vk::AccessFlags2),
                dst: vk::PipelineStageFlags2,
            },
            MakeVisible {
                src: vk::PipelineStageFlags2,
                dst: (vk::PipelineStageFlags2, vk::AccessFlags2),
            },
            MemoryBarrier {
                src: (vk::PipelineStageFlags2, vk::AccessFlags2),
                dst: (vk::PipelineStageFlags2, vk::AccessFlags2),
            },
        }

        impl Display for Barrier {
            fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
                match self {
                    Barrier::Logical => write!(f, "Logical"),
                    Barrier::Execution { .. } => write!(f, "Execution"),
                    Barrier::LayoutTransition { .. } => write!(f, "Layout"),
                    Barrier::MakeAvailable { .. } => write!(f, "MakeAvailable"),
                    Barrier::MakeVisible { .. } => write!(f, "MakeVisible"),
                    Barrier::MemoryBarrier { .. } => write!(f, "MemoryBarrier"),
                }
            }
        }

        let pass_count = self.pass_descs.len() as u32;

        // build graph from references.
        let mut dag = {
            let mut edges = Vec::new();

            intervals.iter().for_each(|(&rid, &(from, to))| {
                #[derive(Clone, Copy, Debug)]
                enum PreviousRef {
                    Write(PassNode, Access),
                    Read(PassNode, Access),
                }
                impl PreviousRef {
                    fn node(&self) -> PassNode {
                        match self {
                            PreviousRef::Write(pass_node, _) => *pass_node,
                            PreviousRef::Read(pass_node, _) => *pass_node,
                        }
                    }
                    fn access(&self) -> Access {
                        match self {
                            PreviousRef::Write(_, access) => *access,
                            PreviousRef::Read(_, access) => *access,
                        }
                    }
                }
                // writes not yet made available
                let mut to_make_available = AccessMask::undefined();
                // writes already made-visible
                let mut made_visible = AccessMask::undefined();
                let mut last_read = Option::<PreviousRef>::None;
                let mut last_write = Option::<PreviousRef>::None;
                let mut last_ref = Option::<PreviousRef>::None;

                // pass, read, write
                let mut current_pass = (from, Access::empty(), Access::empty());
                for pass in PassNode::range_full(from, to, pass_count)
                    .map(|i| PassNode::from_u32(i, pass_count))
                {
                    let mut range = references.range(
                        GraphRef {
                            pass,
                            resource: rid,
                            access: RefAccess::__Min,
                        }..GraphRef {
                            pass,
                            resource: rid,
                            access: RefAccess::__Max,
                        },
                    );

                    while let Some(a) = range.next() {
                        if a.pass != current_pass.0 {
                            if current_pass.1 != Access::empty() {
                                last_read = Some(PreviousRef::Read(current_pass.0, current_pass.1));
                                last_ref = Some(PreviousRef::Read(current_pass.0, current_pass.1));
                            }
                            if current_pass.2 != Access::empty() {
                                last_write =
                                    Some(PreviousRef::Write(current_pass.0, current_pass.2));
                                last_ref = Some(PreviousRef::Write(current_pass.0, current_pass.2));
                            }

                            current_pass = (a.pass, Access::empty(), Access::empty());
                        }

                        // TODO: VkEvents can make this more
                        // fine-grained but also probably have zero
                        // real-world benefit :<

                        match a.access {
                            RefAccess::None => {
                                // make-available previous writes
                                // no-op edge between previous reference and a.pass

                                if let Some(last_ref) = last_ref.as_ref() {
                                    edges
                                        .push(((last_ref.node(), a.pass), (rid, Barrier::Logical)));
                                }
                                // because this is the last node, setting last_ref isn't required.
                            }
                            RefAccess::Read(access) => {
                                // - if read: no writes pending, check for
                                // layout transition, otherwise an edge to
                                // previous write. make sure it is only executed
                                // once.
                                // - if write: make-available writes + make-visible for reads
                                let make_visible_mask = access.into_access_mask() & !made_visible;
                                made_visible = made_visible | make_visible_mask;

                                match last_ref {
                                    None => {}
                                    Some(PreviousRef::Read(pass_node, before)) => {
                                        if let Some(last_write) = last_write.as_ref() {
                                            if !make_visible_mask.is_empty() {
                                                // make-visible reads.
                                                edges.push((
                                                    (last_write.node(), a.pass),
                                                    (
                                                        rid,
                                                        Barrier::MakeVisible {
                                                            src: last_write.access().stage,
                                                            dst: (
                                                                make_visible_mask.stage,
                                                                make_visible_mask.mask,
                                                            ),
                                                        },
                                                    ),
                                                ));
                                            } else {
                                                // still require a after b
                                                edges.push((
                                                    (last_write.node(), a.pass),
                                                    (
                                                        rid,
                                                        Barrier::Execution {
                                                            src: last_write.access().stage,
                                                            dst: access.stage,
                                                        },
                                                    ),
                                                ));
                                            }
                                        }

                                        if before.layout != access.layout {
                                            edges.push((
                                                (pass_node, a.pass),
                                                (
                                                    rid,
                                                    Barrier::LayoutTransition {
                                                        src: (before.stage, before.layout.unwrap()),
                                                        dst: (access.stage, access.layout.unwrap()),
                                                    },
                                                ),
                                            ));
                                        }
                                    }
                                    Some(PreviousRef::Write(write, before)) => {
                                        // make writes visible
                                        if !make_visible_mask.is_empty() {
                                            edges.push((
                                                (write, a.pass),
                                                (
                                                    rid,
                                                    Barrier::MakeVisible {
                                                        src: before.stage,
                                                        dst: (
                                                            make_visible_mask.stage,
                                                            make_visible_mask.mask,
                                                        ),
                                                    },
                                                ),
                                            ));
                                        }
                                        // make all writes available
                                        if !to_make_available.is_empty() {
                                            tracing::debug!(
                                                "making available {:?} for {:?} on {:?}",
                                                to_make_available,
                                                rid,
                                                a.pass
                                            );
                                            edges.push((
                                                (write, a.pass),
                                                (
                                                    rid,
                                                    Barrier::MakeAvailable {
                                                        src: (
                                                            to_make_available.stage,
                                                            to_make_available.mask,
                                                        ),
                                                        dst: access.stage,
                                                    },
                                                ),
                                            ));
                                            // mark that we've made all pending writes available
                                            to_make_available = AccessMask::undefined();
                                        }

                                        if make_visible_mask.is_empty()
                                            && to_make_available.is_empty()
                                        {
                                            // still require a after b
                                            edges.push((
                                                (write, a.pass),
                                                (
                                                    rid,
                                                    Barrier::Execution {
                                                        src: before.stage,
                                                        dst: access.stage,
                                                    },
                                                ),
                                            ));
                                        }
                                    }
                                }
                                current_pass.1 = current_pass.1 | access;
                            }
                            RefAccess::Write(access) => {
                                // - if read: execution barrier against write-after-read
                                // - if write: check for layout transition, otherwise a no-op edge.
                                to_make_available = to_make_available | access.into_access_mask();

                                match last_ref {
                                    None => {}
                                    Some(PreviousRef::Read(pass_node, before)) => {
                                        // execution barrier to ward against write-after-read
                                        edges.push((
                                            (pass_node, a.pass),
                                            (
                                                rid,
                                                Barrier::Execution {
                                                    src: before.stage,
                                                    dst: access.stage,
                                                },
                                            ),
                                        ));
                                    }
                                    Some(PreviousRef::Write(pass_node, before)) => {
                                        if before.layout != access.layout {
                                            // as far as I understand the spec,
                                            // this already makes-available
                                            edges.push((
                                                (pass_node, a.pass),
                                                (
                                                    rid,
                                                    Barrier::LayoutTransition {
                                                        src: (before.stage, before.layout.unwrap()),
                                                        dst: (access.stage, access.layout.unwrap()),
                                                    },
                                                ),
                                            ));
                                        }
                                        // write_no_sync: pass tells us that
                                        // writes do not interleave.
                                        if let Some(last_read) = last_read.as_ref() {
                                            edges.push((
                                                (last_read.node(), a.pass),
                                                (rid, Barrier::Logical),
                                            ));
                                        }
                                    }
                                }

                                current_pass.2 = current_pass.2 | access;
                            }
                            _ => unreachable!(),
                        };
                    }
                }
            });

            let mut dag = petgraph::graph::DiGraph::new();
            let root = dag.add_node(PassNode::First);
            let output = dag.add_node(PassNode::Last);
            for i in 0..self.pass_descs.len() {
                dag.add_node(PassNode::Pass(i));
            }

            for ((from, to), weight) in edges {
                dag.add_edge(
                    from.into_node_idx().into(),
                    to.into_node_idx().into(),
                    weight,
                );
            }

            loop {
                let sinks = dag
                    .externals(petgraph::Direction::Outgoing)
                    .filter(|idx| dag.node_weight(*idx) != Some(&PassNode::Last))
                    .collect::<Vec<_>>();
                if sinks.is_empty() {
                    break;
                }
                for sink in sinks {
                    dag.remove_node(sink);
                }
            }

            #[cfg(any(debug_assertions, test))]
            std::fs::write(
                "render_graph2.dot",
                &format!(
                    "{:?}",
                    petgraph::dot::Dot::with_attr_getters(
                        &dag,
                        &[],
                        &|_graph, edgeref| {
                            format!(
                                "label = \"{},{:#?}\"",
                                edgeref.weight().0.as_u32(),
                                edgeref.weight().1,
                            )
                        },
                        &|_graph, noderef| { format!("label = \"Pass({:?})\"", noderef.weight()) }
                    )
                ),
            )
            .expect("writing render_graph repr");

            dag
        };

        let now = std::time::Instant::now();
        let mut dag = petgraph::stable_graph::StableDiGraph::new();

        let root = dag.add_node(PassNode::First);
        let mut last_write: BTreeMap<GraphResourceId, (NodeIndex, Access)> = self
            .resources
            .keys()
            .filter_map(|id| self.accesses.get(id).map(|access| (*id, (root, *access))))
            .collect::<BTreeMap<_, _>>();

        let mut last_read: BTreeMap<GraphResourceId, (NodeIndex, Access)> = BTreeMap::new();

        // TODO: rewrite finding edges properly.
        // finding out if this graph is cyclical is actually non-trivial
        // some pass might require both a read of a resource 1, and a read of a resource 2, where 2 is the product of another pass writing to resource 1.
        // this could be resolved by copying resource 1 before the write pass.
        // tl;dr: write-after-read makes this all more complicated

        // insert edges between write->read edges of 2 passes
        for (i, pass) in self.pass_descs.iter().enumerate() {
            let node = dag.add_node(PassNode::Pass(i));

            let mut read_accesses = BTreeMap::new();
            for (rid, access) in &pass.reads {
                read_accesses
                    .entry(*rid)
                    .and_modify(|a| {
                        // a single pass must not read one image twice with different layouts
                        *a = *a | *access;
                    })
                    .or_insert(*access);
            }

            for (rid, after) in read_accesses {
                if let Some(&(other, before)) = last_write.get(&rid) {
                    tracing::trace!("adding edge between {other:?} and {node:?} for {rid:?} with ({before:?} -> {after:?})");
                    dag.add_edge(other, node, (rid, (before, after)));
                }
                last_read.insert(rid, (node, after));
            }

            let mut write_accesses = BTreeMap::new();
            for (rid, access) in &pass.writes {
                write_accesses
                    .entry(*rid)
                    .and_modify(|a| {
                        // a single pass must not write one image twice with different layouts
                        *a = *a | *access;
                    })
                    .or_insert(*access);
            }

            for (rid, after) in write_accesses {
                last_write.insert(rid, (node, after));
                if let Some(&(other, read)) = last_read.get(&rid)
                    && other != node
                {
                    tracing::trace!("adding edge between {other:?} and {node:?} for {rid:?} with ({read:?} -> {after:?}) (WaR)");
                    dag.add_edge(other, node, (rid, (read, after)));
                }
            }
        }

        // pseudo pass for tracking outputs
        let output = dag.add_node(PassNode::Last);
        for (id, (node, access)) in self
            .outputs
            .iter()
            .filter_map(|&id| last_write.get(&id).cloned().map(|node| (id, node)))
        {
            dag.add_edge(
                node,
                output,
                (
                    id,
                    (
                        access,
                        // make output writes available
                        Access {
                            stage: vk::PipelineStageFlags2::NONE,
                            mask: vk::AccessFlags2::empty(),
                            ..access
                        },
                    ),
                ),
            );
        }

        // prune dead nodes
        loop {
            let sinks = dag
                .externals(petgraph::Direction::Outgoing)
                .filter(|idx| idx != &output)
                .collect::<Vec<_>>();
            if sinks.is_empty() {
                break;
            }
            for sink in sinks {
                dag.remove_node(sink);
            }
        }

        // handle layout additional transitions
        let edges = dag
            .node_references()
            .map(|(source, _)| {
                let mut per_resourcelayout_multimap: BTreeMap<
                    (GraphResourceId, Option<vk::ImageLayout>),
                    Vec<(Access, NodeIndex)>,
                > = BTreeMap::new();
                let mut resources = BTreeSet::new();

                dag.edges_directed(source, petgraph::Direction::Outgoing)
                    .for_each(|edge| {
                        let (rid, (_, after)) = edge.weight();
                        let target = edge.target();
                        let key = (*rid, after.layout);
                        let item = (*after, target);
                        resources.insert(*rid);

                        per_resourcelayout_multimap
                            .entry(key)
                            .and_modify(|list| list.push(item))
                            .or_insert(vec![item]);
                    });

                let mut edges = vec![];
                for resource in resources {
                    for (a, b) in per_resourcelayout_multimap
                        .range(
                            (resource, None)
                                ..=(resource, Some(vk::ImageLayout::from_raw(i32::MAX))),
                        )
                        .tuple_windows()
                    {
                        let a = a.1;
                        let b = b.1;

                        // create new edge between all members of (a) and (b).
                        // topological mapping will fold all transitions into one.
                        for i in 0..a.len().max(b.len()) {
                            let from = a.get(i).unwrap_or(a.last().unwrap());
                            let to = b.get(i).unwrap_or(b.last().unwrap());

                            let edge = ((from.1, to.1), (resource, (from.0, to.0)));
                            edges.push(edge);
                        }
                    }
                }

                edges
            })
            .flatten()
            .collect::<Vec<_>>();

        for ((from, to), edge) in edges {
            tracing::trace!(
                "adding additional edge between {from:?} and {to:?} for {:?} with ({:?} -> {:?})",
                edge.0,
                edge.1 .0,
                edge.1 .1
            );
            dag.add_edge(from, to, edge);
        }

        // #[cfg(any(debug_assertions, test))]
        // std::fs::write(
        //     "render_graph.dot",
        //     &format!(
        //         "{:?}",
        //         petgraph::dot::Dot::with_attr_getters(
        //             &dag,
        //             &[],
        //             &|_graph, edgeref| { format!("label = \"{:?}\"", edgeref.weight()) },
        //             &|_graph, noderef| { format!("label = \"Pass({:?})\"", noderef.weight()) }
        //         )
        //     ),
        // )
        // .expect("writing render_graph repr");

        let mut topological_map = Vec::new();
        let mut top_dag = dag.clone();
        let mut root_barriers = BTreeMap::new();

        // create topological map of DAG from sink to source
        loop {
            let (sinks, passes): (Vec<_>, Vec<_>) = top_dag
                .externals(petgraph::Direction::Outgoing)
                .filter(|&id| id != root)
                .filter_map(|id| top_dag.node_weight(id).cloned().map(|idx| (id, idx)))
                .unzip();

            if sinks.is_empty() {
                break;
            }

            let mut barriers = BTreeMap::new();

            for &sink in &sinks {
                top_dag
                    .edges_directed(sink, petgraph::Direction::Incoming)
                    .for_each(|edge| {
                        let (rid, (before, after)) = edge.weight();

                        // initial access is transitioned at the beginning
                        // this affects imported resources only.
                        if edge.source() == root {
                            &mut root_barriers
                        } else {
                            &mut barriers
                        }
                        .entry(*rid)
                        .and_modify(|(from, to)| {
                            *from = *from | *before;
                            *to = *to | *after;
                        })
                        .or_insert((*before, *after));
                    });
                top_dag.remove_node(sink);
            }

            let passes = passes
                .into_iter()
                .filter_map(|pass| {
                    if let PassNode::Pass(i) = pass {
                        Some(i)
                    } else {
                        None
                    }
                })
                .map(|i| core::mem::take(&mut self.pass_descs[i]))
                .collect::<Vec<_>>();

            topological_map.push((passes, barriers));
        }
        topological_map.push((vec![], root_barriers));

        //tracing::debug!("mapping: {topological_map:#?}");
        // I don't think this can currently happen with the way passes are added.
        top_dag.remove_node(root);
        if top_dag.node_count() > 0 {
            eprintln!("dag: {top_dag:?}");
            panic!("dag is cyclic!");
        }

        tracing::debug!(
            "resolving render graph: {}ms",
            now.elapsed().as_micros() as f32 / 1e3
        );

        let now = std::time::Instant::now();
        let pool =
            commands::SingleUseCommandPool::new(device.clone(), device.graphics_queue().clone())?;

        let resources = &self.resources;
        let cmds = topological_map
            .into_iter()
            .rev()
            .map({
                |(passes, accesses)| {
                    let cmd = pool.alloc()?;
                    // transitions
                    for (&id, &(from, to)) in accesses.iter() {
                        Self::transition_resource(
                            resources.get(&id).unwrap(),
                            device.dev(),
                            unsafe { &cmd.buffer() },
                            from,
                            to,
                        );
                    }

                    let ctx = RenderContext {
                        device: device.clone(),
                        cmd,
                        resources,
                    };

                    for pass in passes {
                        (pass.record)(&ctx)?;
                    }

                    ctx.cmd.end()?;
                    crate::Result::Ok(ctx.cmd)
                }
            })
            .collect::<crate::Result<Vec<_>>>()?;

        let cmd_list = commands::CommandList(cmds);

        tracing::debug!(
            "recording render graph: {}ms",
            now.elapsed().as_micros() as f32 / 1e3
        );

        Ok(WithLifetime::new(cmd_list))
    }

    pub fn get_outputs(&mut self) -> BTreeMap<GraphResourceId, GraphResource> {
        let outputs = self
            .outputs
            .iter()
            .filter_map(|id| self.resources.remove(id).map(|res| (*id, res)))
            .collect::<BTreeMap<_, _>>();

        outputs
    }

    pub fn transition_resource(
        res: &GraphResource,
        dev: &ash::Device,
        cmd: &vk::CommandBuffer,
        from: Access,
        to: Access,
    ) {
        let barrier: Barrier = match res {
            GraphResource::Framebuffer(arc) => {
                image_barrier(arc.image.handle(), arc.image.format(), from, to, None).into()
            }
            GraphResource::ImportedImage(arc) => {
                image_barrier(arc.handle(), arc.format(), from, to, None).into()
            }
            GraphResource::ImportedBuffer(arc) => {
                buffer_barrier(arc.handle(), 0, arc.len(), from, to, None).into()
            }
            GraphResource::Image(image) => {
                image_barrier(image.handle(), image.format(), from, to, None).into()
            }
            GraphResource::Buffer(buffer) => {
                buffer_barrier(buffer.handle(), 0, buffer.len(), from, to, None).into()
            }
        };

        unsafe {
            dev.cmd_pipeline_barrier2(*cmd, &((&barrier).into()));
        }
    }

    fn transition_resource_to(
        accesses: &mut BTreeMap<GraphResourceId, Access>,
        resources: &BTreeMap<GraphResourceId, GraphResource>,
        dev: &ash::Device,
        cmd: &vk::CommandBuffer,
        id: GraphResourceId,
        to: Access,
    ) {
        let old_access = accesses.get(&id);
        let res = resources.get(&id);

        if let (Some(&old_access), Some(res)) = (old_access, res) {
            Self::transition_resource(res, dev, cmd, old_access, to);
            accesses.insert(id, to);
        }
    }
}

pub enum Barrier {
    Image(vk::ImageMemoryBarrier2<'static>),
    Buffer(vk::BufferMemoryBarrier2<'static>),
}

impl<'a> From<&'a Barrier> for vk::DependencyInfo<'a> {
    fn from(value: &'a Barrier) -> Self {
        let info = vk::DependencyInfo::default();
        let info = match value {
            Barrier::Image(barrier) => info.image_memory_barriers(core::slice::from_ref(barrier)),
            Barrier::Buffer(barrier) => info.buffer_memory_barriers(core::slice::from_ref(barrier)),
        };

        info
    }
}

impl From<vk::ImageMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::ImageMemoryBarrier2<'static>) -> Self {
        Self::Image(value)
    }
}
impl From<vk::BufferMemoryBarrier2<'static>> for Barrier {
    fn from(value: vk::BufferMemoryBarrier2<'static>) -> Self {
        Self::Buffer(value)
    }
}

pub fn buffer_barrier(
    buffer: vk::Buffer,
    offset: u64,
    size: u64,
    before: Access,
    after: Access,
    queue_families: Option<(u32, u32)>,
) -> vk::BufferMemoryBarrier2<'static> {
    vk::BufferMemoryBarrier2::default()
        .buffer(buffer)
        .offset(offset)
        .size(size)
        .src_access_mask(before.mask)
        .src_stage_mask(before.stage)
        .dst_access_mask(after.mask)
        .dst_stage_mask(after.stage)
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

pub fn image_barrier(
    image: vk::Image,
    format: vk::Format,
    before_access: Access,
    after_access: Access,
    queue_families: Option<(u32, u32)>,
) -> vk::ImageMemoryBarrier2<'static> {
    vk::ImageMemoryBarrier2::default()
        .src_access_mask(before_access.mask)
        .src_stage_mask(before_access.stage)
        .dst_access_mask(after_access.mask)
        .dst_stage_mask(after_access.stage)
        .image(image)
        .old_layout(before_access.layout.unwrap_or_default())
        .new_layout(after_access.layout.unwrap_or_default())
        .subresource_range(vk::ImageSubresourceRange {
            aspect_mask: util::image_aspect_from_format(format),
            ..images::SUBRESOURCERANGE_ALL
        })
        .src_queue_family_index(
            queue_families
                .map(|(src, _)| src)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
        .dst_queue_family_index(
            queue_families
                .map(|(_, dst)| dst)
                .unwrap_or(vk::QUEUE_FAMILY_IGNORED),
        )
}

// #[cfg(test)]
// mod tests {
//     use super::*;

//     macro_rules! def_dummy_pass {
//         ($name:ident: {$queue:path, $layout_in:path, $layout_out:path}) => {
//             #[derive(Debug, Clone)]
//             struct $name(Vec<RenderGraphResourceId>, Vec<RenderGraphResourceId>);
//             impl Pass for $name {
//                 fn get_read_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
//                     ResourceAccess::default()
//                 }
//                 fn get_write_resource_access(&self, _id: RenderGraphResourceId) -> ResourceAccess {
//                     ResourceAccess::default()
//                 }

//                 fn get_queue_capability_requirements(&self) -> device::QueueFlags {
//                     $queue
//                 }

//                 fn get_read_dependencies<'a>(
//                     &'a self,
//                 ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
//                     Box::new(self.0.iter().cloned())
//                 }

//                 fn get_write_dependencies<'a>(
//                     &'a self,
//                 ) -> Box<dyn Iterator<Item = RenderGraphResourceId> + 'a> {
//                     Box::new(self.1.iter().cloned())
//                 }

//                 fn record(&self, _ctx: &RenderContext) -> crate::Result<()> {
//                     Ok(())
//                 }
//             }
//         };
//     }

//     def_dummy_pass!(DepthPass: {
//         device::QueueFlags::GRAPHICS,
//         vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL,
//         vk::ImageLayout::DEPTH_ATTACHMENT_OPTIMAL});
//     def_dummy_pass!(RenderPass: {
//         device::QueueFlags::GRAPHICS,
//         vk::ImageLayout::COLORiATTACHMENT_OPTIMAL,
//         vk::ImageLayout::COLOR_ATTACHMENT_OPTIMAL});
//     def_dummy_pass!(AsyncPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::UNDEFINED,
//         vk::ImageLayout::GENERAL});
//     def_dummy_pass!(PostProcessPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::GENERAL,
//         vk::ImageLayout::GENERAL});
//     def_dummy_pass!(PresentPass: {
//         device::QueueFlags::PRESENT,
//         vk::ImageLayout::PRESENT_SRC_KHR,
//         vk::ImageLayout::UNDEFINED});
//     def_dummy_pass!(DepthVisualisationPass: {
//         device::QueueFlags::ASYNC_COMPUTE,
//         vk::ImageLayout::GENERAL,
//         vk::ImageLayout::UNDEFINED});

//     #[test]
//     fn resolve_graph() {
//         let mut graph = RenderGraph::new();
//         let gbuffer = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let depth_image = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let depth_visualisation = graph.add_resource(RenderGraphResourceDesc::Image(ImageDesc {
//             ..Default::default()
//         }));
//         let compute_buffer = graph.add_resource(RenderGraphResourceDesc::Buffer(BufferDesc {
//             ..Default::default()
//         }));

//         graph.add_pass(DepthPass(vec![depth_image], vec![depth_image]));
//         graph.add_pass(DepthVisualisationPass(
//             vec![depth_image, depth_visualisation],
//             vec![depth_visualisation],
//         ));
//         graph.add_pass(AsyncPass(vec![compute_buffer], vec![compute_buffer]));
//         graph.add_pass(RenderPass(
//             vec![depth_image, compute_buffer, gbuffer],
//             vec![gbuffer],
//         ));
//         graph.add_pass(PostProcessPass(vec![gbuffer], vec![gbuffer]));
//         graph.mark_as_output(gbuffer);
//         graph.mark_as_output(depth_image);

//         // graph.resolve();
//     }
// }

pub fn clear_pass(rg: &mut RenderGraph, color: Rgba, target: GraphResourceId) {
    let reads = [(target, Access::transfer_write())].to_vec();
    let writes = [(target, Access::transfer_write())].to_vec();

    let record: Box<RecordFn> = Box::new({
        move |ctx| {
            let target = ctx.get_image(target).unwrap();
            let cmd = &ctx.cmd;

            cmd.clear_color_image(
                target.handle(),
                target.format(),
                vk::ImageLayout::TRANSFER_DST_OPTIMAL,
                color,
                &[images::SUBRESOURCERANGE_COLOR_ALL],
            );
            Ok(())
        }
    });
    rg.add_pass(PassDesc {
        reads,
        writes,
        record,
    });
}

pub fn present_pass(rg: &mut RenderGraph, target: GraphResourceId) {
    let record: Box<RecordFn> = Box::new(|_| Ok(()));
    let reads = vec![(target, Access::present())];
    let writes = vec![(target, Access::present())];
    rg.add_pass(PassDesc {
        reads,
        writes,
        record,
    });
}
